/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "stdafx.h"
#ifdef _USRDLL
#define _QUBFITS_LIB_
#endif 

#include "QUB_LMFitter.h"
#include "lmmin.h"

static char param_names[][16]
 = {"max iterations",
	"epsilon",
	"step bound",
	"ftol",
	"xtol",
	"gtol"};
#define I_MAXITER            0
#define I_EPSILON            1
#define I_STEPBOUND          2
#define I_FTOL               3
#define I_XTOL               4
#define I_GTOL               5
#define QUB_LMFITTER_NPAR    6

#define QUB_LMFITTER_DELTAPCT 0.01


void lm_evaluate_qub( double* par, int m_dat, double* fvec, 
                      void *data, int *info )
{
	((QUB_LMFitter*)(data))->evaluate(par, fvec, info);
}

void lm_print_qub( int n_par, double* par, int m_dat, double* fvec, 
                   void *data, int iflag, int iter, int nfev )
{
	((QUB_LMFitter*)(data))->print(par, fvec, iflag, iter, nfev);
}


QUB_LMFitter::QUB_LMFitter()
: state( QUB_Tree::Create("Fitter") ),
  params(QUB_LMFITTER_NPAR)
{
	params[I_MAXITER] = 100;
	params[I_EPSILON] = 1.0e-14;
	params[I_STEPBOUND] = 100.0;
	params[I_FTOL] = 1.0e-14;
	params[I_XTOL] = 1.0e-14;
	params[I_GTOL] = 1.0e-14;

	state["Name"].setData( getName() );
	QUB_Tree pars = state["Params"];
	for (int i=0; i<QUB_LMFITTER_NPAR; ++i)
		pars[param_names[i]].setData(QTR_TYPE_DOUBLE, params[i]);
}
void __cdecl QUB_LMFitter::Free()
{
	delete this;
}
IQUB_Evts * __cdecl QUB_LMFitter::Events()
{
	return &events;
}
const char * __cdecl QUB_LMFitter::getName()
{
	return "Levenburg-Marquardt";
}
int __cdecl QUB_LMFitter::getParamCount()
{
	return QUB_LMFITTER_NPAR;
}
const char * __cdecl QUB_LMFitter::getParamName(int ix)
{
	if ( (ix >= 0) && (ix < QUB_LMFITTER_NPAR) )
		return param_names[ix];
	return "";
}
double __cdecl QUB_LMFitter::getParamVal(int ix)
{
	if ( (ix >= 0) && (ix < QUB_LMFITTER_NPAR) )
		return params[ix];
	return 0.0;
}
void __cdecl QUB_LMFitter::setParamVal(int ix, double val)
{
	if ( (ix >= 0) && (ix < QUB_LMFITTER_NPAR) ) {
		if ( params[ix] != val ) {
			params[ix] = val;
			state["Params"][param_names[ix]].setData(QTR_TYPE_DOUBLE, val);
			events.send("Fitter ParamVal", this);
		}
	}
}
void QUB_LMFitter::setupData(QF_Data **adata, int ndata)
{
  for (int i=0; i<data.size(); ++i) {
    QF_Data *d = data[i];
    for (int j=0; j<d->nv; ++j)
      delete [] d->vname[j];
    delete [] d->vname;
    delete [] d->vvv;
    delete d;
  }
  data.clear();
  this->fff.resize(ndata);
  this->n = 0;
  for (int i=0; i<ndata; ++i) {
    QF_Data *d = new QF_Data;
    QF_Data *ad = adata[i];
    d->n = ad->n;
    // in actual usage so far, the xx, yy, ww, vvv[i] arrays are persistent though the QF_Data*, vname and vvv are not:
    d->xx = ad->xx;
    d->yy = ad->yy;
    d->ww = ad->ww;
    d->nv = ad->nv;
    d->vname = new char*[d->nv];
    d->vvv = new double*[d->nv];
    for (int j=0; j<d->nv; ++j) {
      int namelen = strlen(ad->vname[j]);
      d->vname[j] = new char[namelen+1];
      memcpy(d->vname[j], ad->vname[j], namelen);
      d->vname[j][namelen] = 0;
      d->vvv[j] = ad->vvv[j];
    }
    data.push_back(d);
    fff[i].resize(d->n);
    this->n += adata[i]->n;
  }
}
int __cdecl QUB_LMFitter::Fit(IQUB_Curve* curve, QF_Data **adata, int ndata)
{
	this->curve = curve;
	setupData(adata, ndata);
	this->iter = 0;
	this->msg = "";
	this->stopRequested = false;

        lm_control_type control;
	control.maxcall = (int) params[I_MAXITER];
	control.stepbound = params[I_STEPBOUND];
	control.epsilon = params[I_EPSILON];
	control.ftol = params[I_FTOL];
	control.xtol = params[I_XTOL];
	control.gtol = params[I_GTOL];

	int ncurvepar = curve->getParamCount();
	useCpar.clear();
	for (int i=0; i<ncurvepar; ++i)
		if ( ! curve->getParamConst(i) )
			useCpar.push_back(i);
	ncurvepar = useCpar.size();

	vector<double> curvepars( ncurvepar );
	control.curvature = new double*[ncurvepar];

	for (int i=0; i<ncurvepar; ++i) {
		curvepars[i] = curve->getParamVal(useCpar[i]);
		control.curvature[i] = new double[ncurvepar];
	}

	lm_minimize( n, ncurvepar, &(curvepars[0]), lm_evaluate_qub, lm_print_qub,
                 this, &control );

	for (int i=ncurvepar-1; i>=0; --i)
		curve->setParamVal(useCpar[i], curvepars[i], (i==0)?1:0);

	errors.clear();
	errors.resize(curve->getParamCount()); // all zeros right?
	for (int i=0; i<ncurvepar; ++i)
		errors[useCpar[i]] = _finite(control.curvature[i][i])
							 ? sqrt(fabs(control.curvature[i][i]))
							 : 0.0;

	for (int i=0; i<ncurvepar; ++i)
		delete [] control.curvature[i];
	delete [] control.curvature;
#if defined(_WIN32)
	unsigned int fstat = _clearfp(); // just in case
	_fpreset();
#endif
	ostringstream ost;
	ost << iter << " iterations: " << lm_shortmsg[control.info] << endl;
	ost << lm_infmsg[control.info];
	msg = ost.str();
	events.send("Fitter Msg", this);

	this->stopRequested = false;
	return 0;
}
void QUB_LMFitter::evaluate(double* par, double* fvec, int *info)
{
	if ( stopRequested )
		*info = -1;
	int ncurvepar = useCpar.size();
	for (int i=ncurvepar-1; i>=0; --i)
		curve->setParamVal(useCpar[i], par[i], (i==0)?1:0);
	int at = 0;
	for (int d=0; d<data.size(); ++d) {
		curve->evalArrayAll(data[d], &(fff[d][0]));
		for (int i=data[d]->n - 1; i>=0; --i)
			fvec[at+i] = data[d]->ww[i] * (data[d]->yy[i] - fff[d][i]);
		at += data[d]->n;
	}
	if ( stopRequested )
		*info = -1;
}
void QUB_LMFitter::print(double* par, double* fvec, int iflag, int iter, int nfev)
{
	// not sure if the inner loop overshoots, leaving bad param vals?
	int ncurvepar = useCpar.size();
	for (int i=ncurvepar-1; i>=0; --i)
		curve->setParamVal(useCpar[i], par[i], (i==0)?1:0);
	for (int i=0; i<data.size(); ++i)
		curve->evalArrayAll(data[i], &(fff[i][0]));

	if ( iter )
		this->iter = iter;
	events.send("Fitter Iter", this);
}
void __cdecl QUB_LMFitter::Stop()
{
	stopRequested = true;
}
int __cdecl QUB_LMFitter::getIterations()
{
	return iter;
}
double __cdecl QUB_LMFitter::getParamErr(int ix)
{
	if ( (errors.size() > ix) && (errors[ix] > 0.0) )
		return errors[ix];
	else if ( data.size() && &(fff[0][0]) )
		// technically wrong for global fits but shouldn't be needed:
		return ChiSqrUncertainty(curve, data[0], &(fff[0][0]), ix, QUB_LMFITTER_DELTAPCT);
	else
		return 0.0;
}
double * __cdecl QUB_LMFitter::getCurve()
{
	// technically wrong for global fits but shouldn't be needed:
	return &(fff[0][0]);
}
const char * __cdecl QUB_LMFitter::getMessage()
{
	return msg.c_str();
}
QTR_Impl* __cdecl QUB_LMFitter::getState()
{
	QTR_Impl* pstate = state.getImpl();
	QTR_INCREF(pstate);
	return pstate;
}
void __cdecl QUB_LMFitter::updateState(QTR_Impl* pustate)
{
	QUB_Tree ustate(pustate);
	if ( ustate["Name"].dataAsString() != this->getName() )
		return;
	QUB_Tree upars = ustate["Params"];
	for (int i=0; i<QUB_LMFITTER_NPAR; ++i) {
		QUB_Tree par = *(upars.find(param_names[i]));
		if ( par.dataCount() > 0 )
			this->setParamVal(i, par.dataAsDouble());
	}
}

IQUB_Fitter* __stdcall CreateLMFitter()
{
	return new QUB_LMFitter;
}


